/*	$Id: codawrap.c,v 1.43 2005/08/25 20:22:51 phil Exp $ */

/* BLURB gpl

                           Coda File System
                              Release 6

          Copyright (c) 1987-2005 Carnegie Mellon University
                  Additional copyrights listed below

This  code  is  distributed "AS IS" without warranty of any kind under
the terms of the GNU General Public Licence Version 2, as shown in the
file  LICENSE.  The  technical and financial  contributors to Coda are
listed in the file CREDITS.

*/


//
// codawrap
// OSR call interface

#include <coda_types.h>
#include <coda.h>
#include <coda_nt.h>
#include <cnode.h>

// Access to root file handle.

static coda_filehandle *root_fh;
static int root_init = FALSE;

// Wrapper calls

NTSTATUS coda_mount(PVOID MediaHandle, PFS_VOL_HANDLE CodaVolumep,
		    PFS_FILE_HANDLE CodaRootDirHandle)
{
    OW_PSEUDO_MOUNT_INFO *info = MediaHandle;
    coda_inode *codaroot;
    PWCHAR ourname = NULL;  /* XXX is this the drive letter? */
    int len; 
    int rc;
    CodaFid fid;

    coda_inode      *ctl_inode;
    coda_filehandle *ctl_fh;
    CodaFid          ctl_fid = CTL_FID;

    NTSTATUS status;

#ifdef DBG
    clean_msg_list();
#endif
    
    ENTRY(D_WRAPPER, "coda_mount");
    
    if ( CODA_MOUNTED(&coda_nt_volume) ) 
	return STATUS_UNSUCCESSFUL;
    
    if (! info ) {
	CDEBUG((D_WRAPPER, "coda_mount: NULL info!\n"));
	return STATUS_UNSUCCESSFUL;
    }

    // Save the coda device object

    coda_device_object = (PDEVICE_OBJECT) *CodaVolumep;
    
    // Build the root inode and file handle ...

    rc = venus_rootfid(&fid);
    if ( rc ) {
	CDEBUG((D_WRAPPER, "coda_mount: Venus error: %d\n", rc));
	return STATUS_UNRECOGNIZED_VOLUME;
    }
    
    coda_nt_volume.v_refcount = 0;

    codaroot = inode_new(&fid, NULL, "\\");
    if (!codaroot) 
	return STATUS_INSUFFICIENT_RESOURCES;

    inode_pin (codaroot);
    root_fh = get_cfh (codaroot, codaroot);
    if (!root_fh)
	return STATUS_INSUFFICIENT_RESOURCES;
	
    CDEBUG((D_WRAPPER, "coda_mount: allocated coda_inode! fh %d\n",
	    root_fh->fh_unique));
    
    rc = coda_getattr(root_fh, &codaroot->c_fattr);

    if ( rc ) {
	CDEBUG((D_WRAPPER, "coda_mount: error %d getting attributes.\n", rc));
	return STATUS_UNRECOGNIZED_VOLUME;
    }

    ctl_inode = inode_new (&ctl_fid, codaroot, ".CONTROL");
    if (!ctl_inode)
	return STATUS_INSUFFICIENT_RESOURCES;

    inode_pin (ctl_inode);
    ctl_fh = get_cfh (ctl_inode, codaroot);
    if (!ctl_fh)
	return STATUS_INSUFFICIENT_RESOURCES;

    // Set the "attributes" on the .CONTROL file.
    
    RtlSecondsSince1970ToTime(100000000, &ctl_inode->c_fattr.CreationTime);
    RtlSecondsSince1970ToTime(100000000, &ctl_inode->c_fattr.LastAccessTime);
    RtlSecondsSince1970ToTime(100000000, &ctl_inode->c_fattr.LastModifiedTime);
    ctl_inode->c_fattr.ValidDataLength.QuadPart = 0;
    ctl_inode->c_fattr.DiskLength.QuadPart = 0;
    ctl_inode->c_fattr.Attributes = FILE_ATTRIBUTE_NORMAL;
    ctl_inode->c_type = C_VREG;
    inode_setflag (ctl_inode, I_VATTR);

    // Set the coda volume information

    coda_nt_volume.v_namelen = 0;
    coda_nt_volume.v_name = NULL;
    coda_nt_volume.v_rootinode = codaroot;
    coda_nt_volume.v_rootfh = root_fh;
    coda_nt_volume.v_ctlinode = ctl_inode;
    coda_nt_volume.v_ctlfh = ctl_fh;
    coda_nt_volume.v_status = V_MOUNTED | V_ROOT;
    
    *CodaRootDirHandle = (FS_FILE_HANDLE) root_fh;
    *CodaVolumep = (FS_VOL_HANDLE) &coda_nt_volume;

    CDEBUG((D_WRAPPER, "coda_mount: Volume is 0x%x\n", &coda_nt_volume));
    
    EXIT(D_WRAPPER, "coda_mount");
    return STATUS_SUCCESS;
}

NTSTATUS coda_unmount(IN FS_VOL_HANDLE CodaVolume)
{
    struct coda_ntvolume *Volume = (struct coda_ntvolume *)CodaVolume;

    DbgPrint ("coda_umount: entry!\n");

    // ENTRY(D_WRAPPER, "coda_unmount");

    if (Volume != &coda_nt_volume) {
      CDEBUG((D_WRAPPER, "coda_unmount: Wrong Volume!\n"));
      // return some error code
      return STATUS_INVALID_PARAMETER_1;
    }
    
    // Clear request queue

    
#ifdef DBG
    clean_msg_list();
#endif

    // Free all objects allocated ..

    inode_freeall();
    
    coda_nt_volume.v_rootinode = NULL;
    coda_nt_volume.v_rootfh  = NULL;
    coda_nt_volume.v_status = 0;

    if (coda_nt_volume.v_refcount != 1) {
      CDEBUG((D_WRAPPER, "coda_unmount: refs=%d!\n",
	      coda_nt_volume.v_refcount));
    }

    EXIT(D_WRAPPER, "coda_unmount");
    return STATUS_SUCCESS;
}


//
//  coda_verify -- is this a Coda Volume ?
//

NTSTATUS coda_verify(IN FS_VOL_HANDLE CodaVolume)
{
    struct coda_ntvolume *Volume = (struct coda_ntvolume *)CodaVolume;

    ENTRY(D_SUPER, "coda_verify");
    
    if (Volume != &coda_nt_volume || !CODA_ROOTVOL(Volume)
	|| !CODA_MOUNTED(Volume)) {
	CDEBUG((D_WRAPPER, "coda_verify: returning wrong volume.\n"));
	CDEBUG((D_WRAPPER, "    :Volume = 0x%x, coda_nt_volume = 0x%x\n",
	       Volume, &coda_nt_volume));
	return STATUS_WRONG_VOLUME;
    }

    EXIT(D_SUPER, "coda_verify");
    return STATUS_SUCCESS;
}


//
// coda_get_volume_attributes
//

NTSTATUS coda_get_volume_attributes(IN FS_VOL_HANDLE CodaVolume,
				    OUT PFS_VOL_ATTRIBUTES Attributes,
				    PULONG BufferSize)
{
    struct coda_ntvolume *Volume = (struct coda_ntvolume *)CodaVolume;
    NTSTATUS Status;
    ANSI_STRING AnsiString;
    UNICODE_STRING UnicodeString;
    struct coda_statfs statfs;
    int rc;

    WCHAR vname[10] = L"Coda";
    
    ENTRY(D_LOOKATTR, "coda_get_volume_attributes");

    if (Volume != &coda_nt_volume) {
	CDEBUG((D_WRAPPER, "coda_get_volume_attributes: Wrong Volume!\n"));
	CDEBUG((D_WRAPPER, "    :Volume = 0x%x, coda_nt_volume = 0x%x\n",
	       Volume, &coda_nt_volume));
	// Return an error?
    }
  
    //
    // If this is a call to get the attributes for the UNC root
    // volume... we just make a something.
    //
    if( ! CODA_ROOTVOL(Volume) ) {
	DbgPrint ("coda_get_volume_attributes: Non-Root volume!\n");
    }

    
    //
    //  Allocate attributes with space for "coda" label.
    // Attributes = coda_malloc (sizeof(FS_VOL_ATTRIBUTES) + 10 * sizeof(WCHAR));
    
    //
    // Clear the attributes are we are returning.
    //
    RtlZeroMemory(Attributes, sizeof(FS_VOL_ATTRIBUTES) + 10 * sizeof(WCHAR));

    // 
    // Get the cache statistics from venus
    //

    rc = venus_statfs (&statfs);
    if (rc) {
	Attributes->VolumeSize.HighPart = 0;
	Attributes->VolumeSize.LowPart = 72727272;
	Attributes->FreeSpace.HighPart = 0;
	Attributes->FreeSpace.LowPart = 72727272;
    } else {
	Attributes->VolumeSize.HighPart = 0;
	Attributes->VolumeSize.LowPart = statfs.f_blocks;
	Attributes->FreeSpace.HighPart = 0;
	Attributes->FreeSpace.LowPart = statfs.f_bavail;
    }
    
    Attributes->CreationTime = Volume->v_rootinode->c_fattr.CreationTime;
    Attributes->BytesPerAllocationUnit = 1024;
    Attributes->DeviceType = FILE_DEVICE_NETWORK_FILE_SYSTEM; // Was _DISK
    Attributes->DeviceCharacteristics =
	FILE_REMOTE_DEVICE | FILE_CASE_SENSITIVE_SEARCH
	| FILE_CASE_PRESERVED_NAMES;

    memcpy(&Attributes->VolumeLabel[0], vname, wcslen(vname) * sizeof(WCHAR));
    Attributes->LabelLength = wcslen(vname) * sizeof(WCHAR);
    
    EXIT(D_WRAPPER, "coda_get_volume_attributes");
    return(STATUS_SUCCESS);
}


//
// coda_release   XXX needs a lot more work!
//

NTSTATUS coda_release(FS_FILE_HANDLE handle)
{
    coda_filehandle *fh = (coda_filehandle *) handle;
    coda_inode *inode = fh->fh_inode;
    struct coda_ntvolume *ntvol;
    NTSTATUS rc =0;
    int flags;
    
    ENTRY(D_LOOKATTR, "coda_release");
    rc = STATUS_SUCCESS;
    
    if (!CODA_ISOURS(inode)) {
	return STATUS_INVALID_PARAMETER_1;
    }

    ntvol = inode->c_ntvol;
    if (!CODA_NTVOL(ntvol)) {
	return STATUS_UNRECOGNIZED_VOLUME;    
    }
    
    CDEBUG((D_LOOKATTR, "coda_release: for %s %s (fh %d%s)\n", inode->c_name,
	    coda_f2s(&inode->c_fid), fh->fh_unique,
	    fh->fh_isopen ? " is open": ""));

    if (fh == coda_nt_volume.v_ctlfh) {
	CDEBUG ((D_LOOKATTR, "coda_release: control fh, returning success\n"));
	return STATUS_SUCCESS;
    }

    LockInode (inode);

    CDEBUG((D_LOOKATTR, "coda_release: freeing fh %d.\n", fh->fh_unique));
    if (fh->fh_isopen) {
	if (!(--(inode->c_open_cnt))) {
	    CDEBUG((D_FILE, "coda_release: Closing inode file object.")); 
	    // Close the object!
	    if (inode->c_object != NULL) {
		inode_close_handle(inode);
	    } else {
		DbgPrint("coda_release: Closing NULL c_object\n");
	    }
	}
	
	// Tell Venus
	rc = venus_close(&inode->c_fid, fh->fh_openflags, fh->fh_uid);
	
	// ignore error
	if ( rc ) {
	    DbgPrint("coda_release: venus_close error %d\n", rc);
	}
	
	// being paranoid ...
	fh->fh_isopen = 0;
    }
    
    if (!inode_hasflag (inode, I_PINNED)) {
	CDEBUG((D_WRAPPER, "coda_release: releasing unpinned inode\n"));
    }

    UnLockInode(inode);

    free_cfh (fh);

    EXIT (D_LOOKATTR, "coda_release");
    return rc;
}

// 
// Local helper routine
//

static void coda_vattr2FATTR(struct coda_vattr *Attributes,
			     PFS_FILE_ATTRIBUTES FileAttributes, PULONG Type)
{
    RtlZeroMemory(FileAttributes, sizeof(FS_FILE_ATTRIBUTES));

    //
    // Do the conversions for the data being send back
    //
    RtlSecondsSince1970ToTime(Attributes->va_atime.ts_sec,
			      &FileAttributes->CreationTime);
    RtlSecondsSince1970ToTime(Attributes->va_atime.ts_sec,
			      &FileAttributes->LastAccessTime);
    RtlSecondsSince1970ToTime(Attributes->va_mtime.ts_sec,
			      &FileAttributes->LastModifiedTime);
    
    //
    // File size
    //
    FileAttributes->ValidDataLength.QuadPart = Attributes->va_size;
    
    //
    // Allocation size
    //
    FileAttributes->DiskLength.QuadPart = Attributes->va_size;
    
    // 
    // File type
    //
    if( Attributes->va_type == C_VREG )
	FileAttributes->Attributes = FILE_ATTRIBUTE_NORMAL;
    
    if(Attributes->va_type == C_VDIR)
	FileAttributes->Attributes = FILE_ATTRIBUTE_DIRECTORY
	    | FILE_ATTRIBUTE_NORMAL;
    
    switch(Attributes->va_type) {
    case C_VDIR:
	*Type = OSRFS_TYPE_DIRECTORY;
	
	CDEBUG((D_LOOKATTR, "coda_vattr2FATTR: File is a directory\n"));
	break;
	
    case C_VLNK:
	*Type = OSRFS_TYPE_SYMBOLIC_LINK;
	
	CDEBUG((D_LOOKATTR, "coda_vattr2FATTR: File is a link\n"));
	break;
	
    case C_VREG:
	*Type = OSRFS_TYPE_FILE;
	
	CDEBUG((D_LOOKATTR, "coda_vattr2FATTR: File is a regular file\n"));
	break;
	
    default :
	*Type = C_VNON;
	CDEBUG((D_LOOKATTR, "coda_vattr2FATTR: unknown file type!\n"));
    }
    
    return;
    
}

//
// coda_internal_getattr using inodes.  (was coda_getattr)
// 

NTSTATUS coda_internal_getattr(IN coda_inode *inode,
			       IN OUT PFS_FILE_ATTRIBUTES FileAttributes)
{
    NTSTATUS rc;
    struct coda_ntvolume *ntvol;
    struct coda_vattr Attributes;

    FS_FILE_ATTRIBUTES FileAttr;

    if (FileAttributes == NULL)
	FileAttributes = &FileAttr;

    CDEBUG ((D_LOOKATTR, "coda_internal_getattr: Starting on %s. %s\n",
	     inode->c_name, coda_f2s(&inode->c_fid)));

    LockInode(inode);
    
    try {

	//
	// Do some sanity checking to be sure that the handle that
	// was passed in is really a valid file handle.
	//
	if (! CODA_ISOURS(inode) )  {
	    CDEBUG((D_WRAPPER, "coda_internal_getattr: NOT A FILE!\n"));
	    return(STATUS_INVALID_PARAMETER_1);
	}
	
	//
	// Zero attributes buffer before anything else.  This pre-sets
	// all fields to zero.  The wrapper wants us to pass back any
	// fields we do not support as zero.
	//
	__try {
	    RtlZeroMemory(FileAttributes, sizeof(FS_FILE_ATTRIBUTES));
	}
	except(EXCEPTION_EXECUTE_HANDLER) {
	    CDEBUG((D_WRAPPER,
		    "coda_getattar: can't clear buffer!\n"));
	    return(STATUS_INVALID_USER_BUFFER);
	}
	
	/* likely we are luck and have this stuff already */
	if (!inode_hasflag(inode, I_VATTR)) {

	    rc = venus_getattr(&inode->c_fid, &Attributes, GetNtUID());
	    
	    if( rc ) {
		CDEBUG((D_LOOKATTR, "coda_internal_getattr: venus_getattr %s "
		       "returns error %d\n", coda_f2s(&inode->c_fid), rc));
		return(rc);
	    }
	    
	    /* fill in return value and store result in coda_inode */
	    coda_vattr2FATTR(&Attributes, &inode->c_fattr, &inode->c_type);
	    inode_setflag (inode, I_VATTR);
	}
	__try {
	    // Return the attributes.
	    RtlCopyMemory(FileAttributes, &inode->c_fattr,
			  sizeof(FS_FILE_ATTRIBUTES));
	}
	except (EXCEPTION_EXECUTE_HANDLER) {
	    CDEBUG((D_WRAPPER, "coda_internal_getattr: can't copy attributes.\n"));
	    return STATUS_INVALID_USER_BUFFER;
	}
	
    } 
    finally {
	UnLockInode(inode);
    }
    
    CDEBUG ((D_LOOKATTR, "coda_internal_getattr: finished with %s\n",
	     coda_f2s(&inode->c_fid)));
    return STATUS_SUCCESS;
}

//
// coda_getattr
//
NTSTATUS coda_getattr(IN FS_FILE_HANDLE FileHandle,	
			       IN OUT PFS_FILE_ATTRIBUTES FileAttributes)
{
    coda_filehandle *fh = (coda_filehandle *)FileHandle;
    coda_inode * inode = fh->fh_inode;
    NTSTATUS rc;

    CDEBUG((D_LOOKATTR, "coda_getattr: Started on %s. (fh %d)\n",
	    inode->c_name, fh->fh_unique));

    rc = coda_internal_getattr (inode, FileAttributes);

    CDEBUG((D_LOOKATTR, "coda_getattr: Exiting. (fh %d)\n",
	    fh->fh_unique));

    return rc;
}

//
// Support for coda_setattr 
//

static void coda_FATTR2vattr(PFS_FILE_ATTRIBUTES FileAttributes,
			     struct coda_vattr *Attributes)
{
    //
    // Convert times ... which is about all we can do to the attrs!
    //

    if (!RtlTimeToSecondsSince1970(&FileAttributes->LastAccessTime,
				   &Attributes->va_atime.ts_sec)) {
	CDEBUG ((D_LOOKATTR,
		 "coda_FATTR2vattr: Failed to convert LastAccessTime\n"));
	Attributes->va_atime.ts_sec = 0;
	Attributes->va_atime.ts_nsec = 0;
    }

    if (!RtlTimeToSecondsSince1970(&FileAttributes->LastModifiedTime,
				   &Attributes->va_mtime.ts_sec)) {
	CDEBUG ((D_LOOKATTR,
		 "coda_FATTR2vattr: Failed to convert LastModifiedTime\n"));
	Attributes->va_mtime.ts_sec = 0;
	Attributes->va_mtime.ts_nsec = 0;
    }
    
    return;
    
}

//
// coda_setattr
//

NTSTATUS coda_setattr(IN FS_FILE_HANDLE FileHandle,
		      IN OUT PFS_FILE_ATTRIBUTES FileAttributes)
{

    NTSTATUS rc;
    coda_filehandle *fh = (coda_filehandle *) FileHandle;
    coda_inode * inode = fh->fh_inode;
    coda_ntvolume *ntvol;
    struct coda_vattr Attributes;
    
    CDEBUG((D_LOOKATTR, "coda_setattr: Started on %s (fh %d).\n",
	    inode->c_name, fh->fh_unique));
    
    
    //
    // Do some sanity checking to be sure that the handle that
    // was passed in is really a valid file handle.
    //
    if (! CODA_ISOURS(inode) )  {
	CDEBUG((D_WRAPPER, "coda_setattr: NOT A FILE!\n"));
	return(STATUS_INVALID_PARAMETER_1);
    }

    LockInode (inode);
    
    // Set the fields ...
    coda_FATTR2vattr(FileAttributes, &Attributes);
    Attributes.va_ctime.ts_nsec = -1;
    Attributes.va_ctime.ts_sec = -1;
    Attributes.va_type = C_VNON;
    Attributes.va_mode = -1;
    Attributes.va_nlink = -1;
    Attributes.va_uid = -1;
    Attributes.va_gid = -1;
    Attributes.va_fileid = -1;
    Attributes.va_size = -1;
    Attributes.va_blocksize = -1;
    Attributes.va_gen = -1;
    Attributes.va_flags = -1;
    Attributes.va_rdev = -1;
    Attributes.va_bytes = -1;
    Attributes.va_filerev = -1;
    
    rc = venus_setattr(&inode->c_fid, &Attributes, fh->fh_uid);
    
    if( rc ) {
	CDEBUG((D_LOOKATTR, "coda_setattr: venus_setattr %s "
		"returns error %d\n", coda_f2s(&inode->c_fid), rc));
	UnLockInode (inode);
	return(rc);
    }
    
    /* Just update the Attar  */
    inode->c_fattr.LastAccessTime = FileAttributes->LastAccessTime;
    inode->c_fattr.LastModifiedTime = FileAttributes->LastModifiedTime;

    UnLockInode (inode);
    
    EXIT(D_LOOKATTR, "coda_setattr");
    return STATUS_SUCCESS;
}


//
// coda_lookup -- implements the FS_LOOKUP wrapper call
//
NTSTATUS coda_internal_lookup(IN coda_inode *dirinode,
			      IN PCHAR FileName,
			      IN BOOLEAN CaseInsensitive,
			      OUT coda_inode **fileinode)
{
    NTSTATUS Status;
    
    struct coda_vattr Attributes;

    coda_inode *inode = NULL;
    CodaFid fid;
    int type;
    
   
    //
    // Null the return parameters, in case we bail-out early
    //
    *fileinode = NULL;
    
  
    //
    // Looks like the input parameters are OK.  Let's do the
    // lookup...
    //
    CDEBUG((D_LOOKATTR, "coda_internal_lookup: Lookup of %s in %s%s\n",
	   FileName, coda_f2s(&dirinode->c_fid),
	   (CaseInsensitive?" Case Insensitive":"")));

    LockInode(dirinode);

    GrabLock();
    inode = inode_find_name(dirinode, FileName, CaseInsensitive);
    DropLock();
    
    if (!inode) {
	CDEBUG((D_LOOKATTR, "coda_internal_lookup: "
	       "cache miss, going to Venus\n"));

	Status = venus_lookup(&dirinode->c_fid,
			      (const char *)FileName, 
			      strlen(FileName),
			      &type, &fid, GetNtUID());
	
	if (Status) {                   
	    CDEBUG((D_LOOKATTR, "coda_internal_lookup: "
		   "venus_lookup failed. Status = 0x%x\n",
		   Status));
	    UnLockInode(dirinode);
	    return STATUS_OBJECT_NAME_NOT_FOUND;
	}

	GrabLock();
	inode =  inode_new(&fid, dirinode, FileName);
	DropLock();
	
	if(!inode) {
	    CDEBUG((D_LOOKATTR, "coda_internal_lookup: inode_new failed!!\n"));
	    UnLockInode(dirinode);
	    return STATUS_INVALID_PARAMETER_1;
	}
    } else {
	CDEBUG((D_LOOKATTR, "coda_internal_lookup: cache hit.\n"));
    }

    if (!inode_hasflag (inode, I_VATTR)) {
	Status = coda_internal_getattr(inode, NULL);
	if ( !NT_SUCCESS(Status) ) {
	    CDEBUG((D_LOOKATTR, "coda_internal_lookup: "
		   "Cannot get attributes of inode %s "
		   "during lookup: %d\n",
		   coda_f2s(&inode->c_fid), Status));
	    GrabLock();
	    inode_release (inode);	
	    DropLock();
	    return STATUS_OBJECT_NAME_NOT_FOUND;
	}
    }

    UnLockInode(dirinode);

    *fileinode = inode;
    
    CDEBUG((D_LOOKATTR, "coda_internal_lookup: returning success\n"));
    return STATUS_SUCCESS;
}


NTSTATUS coda_lookup(IN FS_FILE_HANDLE DirectoryHandle,
		     IN PWCHAR FileName,
		     IN BOOLEAN CaseInsensitive,
		     OUT PFS_FILE_HANDLE FileHandle,
		     OUT PULONG Type)
     
{
    coda_filehandle *dfh = (coda_filehandle *)DirectoryHandle;
    coda_filehandle *nfh;   
    ANSI_STRING AnsiString;
    UNICODE_STRING UnicodeString;
    CHAR NormalFileName[CODA_MAXNAMLEN];

    coda_inode *dirinode = dfh->fh_inode;
    coda_inode *inode;
    
    int rc;

    CDEBUG ((D_LOOKATTR, "coda_lookup: dir fh %d\n", dfh->fh_unique));

    //
    // First, do some sanity checking to be sure that the handle
    // that was passed in is really a valid directory handle.
    // For us, this means that it is a pointer to an inode. (XXX?)
    //
    if (!CODA_ISOURS(dirinode) ) {
	return(STATUS_INVALID_PARAMETER_1);
    }
    
    if (!CODA_ISDIR(dirinode)) {
	return(STATUS_NOT_A_DIRECTORY);
    }

    //
    // Next, convert the wide-character string to a "normal" null
    // terminated c-string. One would think this would be easier.
    //
    RtlInitUnicodeString(&UnicodeString, FileName);
    RtlUnicodeStringToAnsiString(&AnsiString, &UnicodeString, TRUE);
	
    if (AnsiString.Length > CODA_MAXNAMLEN) {
	return(STATUS_INVALID_PARAMETER_2);
    }
	
    memcpy(NormalFileName, AnsiString.Buffer, AnsiString.Length);
    NormalFileName[AnsiString.Length] = 0x00;
    RtlFreeAnsiString(&AnsiString);

    rc = coda_internal_lookup(dfh->fh_inode, NormalFileName,
			      CaseInsensitive, &inode);

    if (rc == STATUS_SUCCESS) {

	if (inode == coda_nt_volume.v_ctlinode) {
	    nfh = coda_nt_volume.v_ctlfh;

	    CDEBUG ((D_LOOKATTR, "coda_lookup: returning control fh %d\n",
		     nfh->fh_unique));
	} else {

	    GrabLock();
	    // Note: get_cfh produces a unique fh for each uid.
	    nfh = get_cfh(inode, dfh->fh_inode);
	    DropLock();
	    
	    if (!nfh)
		return STATUS_INSUFFICIENT_RESOURCES;
	    
	    CDEBUG ((D_LOOKATTR, "coda_lookup: returning fh %d\n",
		     nfh->fh_unique));
	}

	*FileHandle = (FS_FILE_HANDLE) nfh;
	*Type = inode->c_type;
    }

    // Checked for zapped dirs ...

    if (inode_hasflag(dirinode, I_ZAPPED)) {
 	coda_filehandle *fh;
	inode_unsetflag(dirinode, I_ZAPPED);

	fh = dirinode->c_fh_list;
	while (fh != NULL) {
	    CDEBUG((D_DOWNCALL, "coda_lookup: zapping fh %d\n", fh->fh_unique));
	    (void) OwPurgeCache(fh);  // XXX should use ret val
	    fh = fh->fh_next;
	}
    }    

    EXIT(D_LOOKATTR, "coda_lookup");
    return rc;
}

//
// Path lookup  -- XXX make similar to coda_internal_lookup
//

NTSTATUS coda_internal_path_lookup(IN coda_inode *dirinode,
			      IN PCHAR FilePath,
			      IN BOOLEAN CaseInsensitive,
			      OUT coda_inode **fileinode)
{
    NTSTATUS Status;
    
    CHAR NormalFilePath[CODA_MAXNAMLEN];
    CHAR *FileName;
    CHAR *RemainPath;
    CHAR *temp;
    
    coda_inode *inode = NULL;

    ENTRY (D_PIOCTL, "coda_internal_path_lookup"); 

    // If we fall through everything, there was "no" file name,
    // so we return the coda root directory.
    Status = STATUS_SUCCESS;
    inode = coda_nt_volume.v_rootinode; 

    // Get the string in the local space for modification.

    if (strlen(FilePath) > CODA_MAXNAMLEN) {
	return(STATUS_INVALID_PARAMETER_2);
    }
    strcpy(NormalFilePath, FilePath);

    // Make sure all file separators are /.
    temp = NormalFilePath;
    while (*temp) {
	if (*temp == '\\')
	    *temp = '/';
	temp++;
    }

    // Walk the file name ...
    RemainPath = NormalFilePath;
    while (*RemainPath == '/')
	RemainPath++;
    while (*RemainPath) {
	FileName = RemainPath;
	while (*RemainPath && (*RemainPath != '/'))
	    RemainPath++;
	if (*RemainPath == '/') {
	    *RemainPath = 0;
	    RemainPath++;
	    while (*RemainPath == '/')
		RemainPath++;
	}

	CDEBUG( (D_PIOCTL, "coda_internal_path_lookup: "
		"FileName = '%s', RemainPath= '%s'\n", FileName, RemainPath)); 

	Status = coda_internal_lookup (dirinode, FileName, CaseInsensitive,
				       &inode);

	if (Status != STATUS_SUCCESS) {
	    CDEBUG( (D_PIOCTL,
		    "coda_internal_path_lookup: returning error %d\n", Status));
	    return Status;
	}
	dirinode = inode;
    }

    *fileinode = inode;

    EXIT (D_PIOCTL, "coda_internal_path_lookup"); 

    return Status;
}


//
// coda_read
//
//   This routine reads data from the container file.
//
// Inputs:
//   FileHandle - the file being read
//   FileOffset - the first byte of the read
//   Buffer - the buffer where the data should be stored
//   ReadSize - the number of bytes to be read.
//
// Outputs:
//   BytesRead - the number of bytes actually read
//
// Returns:
//   SUCCESS - the operation completed successfully
//
// Notes:
//   None.
//

NTSTATUS coda_read(IN FS_FILE_HANDLE FileHandle,
		   IN LARGE_INTEGER FileOffset,
		   IN PMDL Buffer,
		   IN ULONG ReadSize,
		   OUT PULONG BytesRead)
{
    PIRP irp;
    KEVENT event;
    PIO_STACK_LOCATION ioStackLocation;
    PDEVICE_OBJECT fsdDevice;
    coda_filehandle *fh = (coda_filehandle *) FileHandle;
    coda_inode * inode = fh->fh_inode;
    IO_STATUS_BLOCK iosb;
    
    ENTRY(D_WRAPPER, "coda_read");
    
    //
    // Sanity checking
    //
    if ( !CODA_ISOURS(inode) || fh == coda_nt_volume.v_ctlfh) {
	CDEBUG((D_WRAPPER, "coda_read: Bad file %s\n", inode->c_name));
	return STATUS_INVALID_PARAMETER_1;
    }
    
    CDEBUG((D_WRAPPER, "coda_read: Reading %s (fh %d: size %d).\n",
	    inode->c_name, fh->fh_unique, ReadSize));

    if ( coda_get_fileobject(fh, C_O_READ) ) {
	CDEBUG((D_WRAPPER, "coda_read: coda_get_fileobject error.\n"));
	return STATUS_INVALID_PARAMETER_1;
    }

    
    //
    // Figure out where to send the IRP
    //
    fsdDevice = IoGetRelatedDeviceObject(inode->c_object);
    
    //
    // Set up our event for use.
    //
    KeInitializeEvent(&event, NotificationEvent, FALSE);
    
    //
    // Allocate and build the IRP we'll be sending to the FSD.
    //    
    irp = IoAllocateIrp(fsdDevice->StackSize, FALSE);    
    if (!irp) {        
	//
	// Allocation failed, presumably due to memory allocation
	// failure.
	//        
	return STATUS_INSUFFICIENT_RESOURCES;        
    }
    
    irp->MdlAddress = Buffer;    
    irp->UserEvent = &event;    
    irp->UserIosb = &iosb;    
    irp->Tail.Overlay.Thread = PsGetCurrentThread();    
    irp->Tail.Overlay.OriginalFileObject= inode->c_object;
    irp->RequestorMode = KernelMode;
    
    //
    // Indicate that this is a READ operation.
    //    
    irp->Flags = IRP_READ_OPERATION | IRP_NOCACHE;
    irp->UserBuffer = MmGetMdlVirtualAddress(Buffer);    
    
    //
    // Set up the next I/O stack location.  These are the parameters
    // that will be passed to the underlying driver.
    //
    
    ioStackLocation = IoGetNextIrpStackLocation(irp);    
    ioStackLocation->MajorFunction = IRP_MJ_READ;    
    ioStackLocation->MinorFunction = 0;    
    ioStackLocation->DeviceObject = fsdDevice;    
    ioStackLocation->FileObject = inode->c_object;
    ioStackLocation->Parameters.Read.Length = ReadSize;
    ioStackLocation->Parameters.Read.ByteOffset = FileOffset;
    
    //
    // Use the generic completion routine.  This ensures two things:
    //  (1) the MDL we pass won't get freed;
    //  (2) we get I/O completion without the cost.
    //
    IoSetCompletionRoutine(irp, NullDiscardIrpCompletion, 0, TRUE, TRUE,
			   TRUE);
    
    
    //
    //  I/O operation is built.  Pass it to the underlying file system.
    //
    (void) IoCallDriver(fsdDevice, irp);
    
    
    //
    // Wait for the I/O to complete.
    //
    KeWaitForSingleObject(&event, Executive, KernelMode, TRUE, 0);
    
    
    //
    // Copy results from io status block.
    //
    *BytesRead = iosb.Information;
    
    
    //
    // Done.
    //
    CDEBUG ((D_WRAPPER, "coda_read: exit: %d bytes read, status %x\n",
	     *BytesRead, iosb.Status));
    return iosb.Status;
}

// coda_set_length -- It appears this must be called before
// any call to coda_write ... to make sure there is enough
// room in the file for the write.

NTSTATUS coda_set_length(IN FS_FILE_HANDLE FileHandle,
			 IN LARGE_INTEGER AllocationSize,
			 IN LARGE_INTEGER DataSize)
{
    coda_filehandle *fh = (coda_filehandle *) FileHandle; 
    coda_inode * inode = fh->fh_inode;
    struct coda_vattr f_attr;
    int rc;

    CDEBUG ((D_WRAPPER, "coda_set_length: Allocate %d, DataSize %d\n",
	     AllocationSize.LowPart, DataSize.LowPart));

    if ( !CODA_ISOURS(inode) || fh == coda_nt_volume.v_ctlfh ) {
	CDEBUG((D_WRAPPER, "coda_set_length: Bad file %s\n"));
	return STATUS_INVALID_PARAMETER_1;
    }

    CDEBUG((D_WRAPPER, "coda_set_length: Setting %s. (fh %d)\n",
	    inode->c_name, fh->fh_unique));

    if ( coda_get_fileobject(fh, C_O_WRITE) ) {
	CDEBUG((D_WRAPPER, "coda_set_length: coda_get_fileobject error.\n"));
	return STATUS_INVALID_PARAMETER_1;
    }

    LockInode(inode);

    /* Update Local copy! */
    inode->c_fattr.ValidDataLength = DataSize;
    inode->c_fattr.DiskLength = AllocationSize;
    rc = STATUS_SUCCESS;
    
    UnLockInode(inode);
    CDEBUG ((D_WRAPPER, "coda_set_length: Exiting. (fh %d)\n", fh->fh_unique));
    return rc;
}


//
// coda_write
//
//   This routine writes data to the backing store
//
// Inputs:
//   FileHandle - the file being read
//   FileOffset - the first byte of the read
//   Buffer - the buffer where the data should be stored
//   WriteSize - the number of bytes to be read.
//
// Outputs:
//   BytesWritten - the number of bytes actually written.
//
// Returns:
//   SUCCESS - the operation was successful.
//
// Notes:
//   None.
//
NTSTATUS coda_write(IN FS_FILE_HANDLE FileHandle,
		    IN LARGE_INTEGER FileOffset,
		    IN PMDL Buffer,
		    IN ULONG WriteSize,
		    OUT PULONG BytesWritten)
{
    coda_filehandle *fh = (coda_filehandle *) FileHandle;
    coda_inode * inode = fh->fh_inode;

    PIRP irp;
    KEVENT event;
    PIO_STACK_LOCATION ioStackLocation;
    PDEVICE_OBJECT fsdDevice;
    IO_STATUS_BLOCK iosb;
    
    ENTRY(D_WRAPPER, "coda_write");
    
    //
    // Sanity checking
    //
    if ( !CODA_ISOURS(inode)  || fh == coda_nt_volume.v_ctlfh ) {
	CDEBUG((D_WRAPPER, "coda_write: Bad file\n"));
	return STATUS_INVALID_PARAMETER_1;
    }
    
    CDEBUG((D_WRAPPER, "coda_write: Writing %s (fh %d).\n", inode->c_name,
	    fh->fh_unique));

    if ( coda_get_fileobject(fh, C_O_WRITE) ) {
	CDEBUG((D_WRAPPER, "coda_write: coda_get_fileobject error.\n"));
	return STATUS_INVALID_PARAMETER_1;
    }
    

    //
    // Figure out where to send the IRP
    //
    fsdDevice = IoGetRelatedDeviceObject(inode->c_object);
    
    //
    // Set up our event for use.
    //
    KeInitializeEvent(&event, NotificationEvent, FALSE);
    
    //
    // Allocate and build the IRP we'll be sending to the FSD.
    //    
    irp = IoAllocateIrp(fsdDevice->StackSize, FALSE);    
    if (!irp) {        
	//
	// Allocation failed, presumably due to memory allocation
	// failure.
	//        
	return STATUS_INSUFFICIENT_RESOURCES;        
    }
    
    irp->MdlAddress = Buffer;    
    irp->UserEvent = &event;    
    irp->UserIosb = &iosb;    
    irp->Tail.Overlay.Thread = PsGetCurrentThread();    
    irp->Tail.Overlay.OriginalFileObject= inode->c_object;    
    irp->RequestorMode = KernelMode;
    
    //
    // Indicate that this is a WRITE operation.
    //
    irp->Flags = IRP_WRITE_OPERATION;    
    
    //
    // Set up the next I/O stack location.  These are the parameters
    // that will be passed to the underlying driver.
    //
    ioStackLocation = IoGetNextIrpStackLocation(irp);    
    ioStackLocation->MajorFunction = IRP_MJ_WRITE;    
    ioStackLocation->MinorFunction = 0;    
    ioStackLocation->DeviceObject = fsdDevice;    
    ioStackLocation->FileObject = inode->c_object;
    ioStackLocation->Parameters.Write.Length = WriteSize;    
    ioStackLocation->Parameters.Write.ByteOffset = FileOffset;
    
    //
    // Use the generic completion routine.  This ensures two things:
    //  (1) the MDL we pass won't get freed;
    //  (2) we get I/O completion without the cost.
    //
    IoSetCompletionRoutine(irp, NullDiscardIrpCompletion, 0, TRUE, TRUE,
			   TRUE);
    
    //
    //  I/O operation is built.  Pass it to the underlying file system.
    //
    (void) IoCallDriver(fsdDevice, irp);
    
    //
    // Wait for the I/O to complete.
    //
    KeWaitForSingleObject(&event, Executive, KernelMode, TRUE, 0);
    
    //
    // Copy results from io status block.
    //
    *BytesWritten = iosb.Information;
    
    //
    // Done.
    //

    EXIT(D_WRAPPER, "coda_write");
    return iosb.Status;
}



static int
coda_venusdir_to_fsdir(struct venus_dirent *de, PFS_DIRECTORY_ENTRY Entry)
{
    char *name;
    NTSTATUS status;
    PUCHAR Ptr;
    ULONG Length;
    ANSI_STRING AnsiString;
    UNICODE_STRING UnicodeString;
    
    
    //
    // pre-set the entry size to zero, in case we need to back-out
    // of this entry
    //
    Entry->EntrySize = 0x00;
    
    //
    // Store the remote File ID where we can later get it
    //
    Entry->Unused.LowPart = 0x0;
    
    //
    // We also use the High Part of the Unused entry as an EOF flag
    //
    Entry->Unused.HighPart = 0x00;
    
    //
    // Get the name.  Since NT wants Wide Characters, we do
    // the conversion here.
    //
    name = de->d_name;  // this is 0 terminated, see coda-src/dir/codadir.c
    
    RtlInitAnsiString(&AnsiString, name);
    status = RtlAnsiStringToUnicodeString(&UnicodeString, &AnsiString,
					  TRUE);
    if ( ! NT_SUCCESS(status) ) {
	CDEBUG((D_FILE, "coda_venusdir_to_fsdir: Cannot convert to "
	       "Unicode!\n"));
	return STATUS_INSUFFICIENT_RESOURCES;
    }
    
    memcpy(&Entry->Name[0], UnicodeString.Buffer, UnicodeString.Length);
    
    //
    // Store the size of the entry
    //
    Entry->NameSize = UnicodeString.Length;
    
    RtlFreeUnicodeString(&UnicodeString);
    
    //
    // Return the NFS "cookie" entry as the DirectoryEntryIndex
    //
    Entry->DirectoryEntryIndex = 0;
    
    //
    // And return the size of the entire entry
    //
    Entry->EntrySize = sizeof(FS_DIRECTORY_ENTRY) + Entry->NameSize;
    
    return 0;   
}


// this converts a Coda style directory buffer to one for the WRAPPER.
// - we return the number of entries found if we are successful
// - we retrun a negative error if we can't get the attributes


int
coda_readdir_rewrite(coda_inode *dir, unsigned long *DirectoryOffset,
		     void *in, int insize, void *out, int outsize, int count)
{
    NTSTATUS status;
    struct venus_dirent *de;
    char *inp, *end;
    int i;
    char *name;
    int stringoffset = (int) &(((struct venus_dirent *)(0))->d_name);
    UNICODE_STRING UnicodeName;
    ANSI_STRING AnsiName;
    coda_inode *file;
    unsigned long type;
    PFS_FILE_ATTRIBUTES fattr;
    PFS_DIRECTORY_ENTRY Entry = out;
    PFS_DIRECTORY_ENTRY PrevEntry = Entry;
    
    // go through the directory and convert existing format into fsd format
    inp = (char *)in;
    end = inp + insize;
    
    i=0;
    while (i < count) {
	
	// some entries go across the buffer boundary,
	// detect those first
	if ((char *)inp + stringoffset > (char *)end) {
	    de = NULL;
	    break;
	}
	
	de = (struct venus_dirent *) inp;
	if ((char *)inp + stringoffset + de->d_namlen + 1
	    > (char *)end ) {
	    de = NULL;
	    break;
	}
	
	// we have the entry, 
	inp += de->d_reclen;
	*DirectoryOffset += de->d_reclen;
	
	// but maybe its the closing entry at the end? 
	if ( de->d_namlen == 0 || inp > end ) {
	    *DirectoryOffset = 0;
	    break;
	}
	
	CDEBUG((D_FILE, "coda_readdir_rewrite: Entry %d, name %s, "
	       "reclen %d, de %p, end %p\n",
	       i, de->d_name, de->d_reclen, de, end));
	
	// . and .. are not to be reported, skip to the next one 
	if ( ((de->d_namlen ==1)
	      && (RtlCompareMemory(de->d_name, ".", 1) == 1))
	     || ((de->d_namlen ==2)
		 && (RtlCompareMemory(de->d_name, "..",2) == 2))
	     ) {
	    continue;
	}
	
	
	// the entry is fully available, copy it into the output
	// buffer 
	PrevEntry->LastDirectoryEntry = 1;
	coda_venusdir_to_fsdir(de, Entry);
	
	// now put the details into the Entry, dunno what to do on
	// failure
	status = coda_internal_lookup(dir, de->d_name, 0, &file); 

	if ( !NT_SUCCESS(status) ) {
	    CDEBUG((D_WRAPPER, "coda_readdir_rewrite: "
		   "coda_internal_lookup error %s\n", de->d_name));
	    return -STATUS_NO_SUCH_FILE;
	}
	
	fattr = &file->c_fattr;
	Entry->CreationTime = fattr->CreationTime;
	Entry->LastAccessTime = fattr->LastAccessTime;
	Entry->LastModifiedTime = fattr->LastModifiedTime;
	Entry->ValidDataLength = fattr->ValidDataLength;
	Entry->DiskLength = fattr->DiskLength;
	Entry->Attributes = fattr->Attributes;
	Entry->DirectoryEntryIndex = (ULONG) &de;
	
	PrevEntry = Entry;
	Entry = (PFS_DIRECTORY_ENTRY)
	    ((char *)Entry + Entry->EntrySize);
	i++;
    }
    
    EXIT(D_WRAPPER, "coda_readdir_rewrite");
    return i;
}


//
// coda_readdir
//
// - we set *DirectoryOffset to 0 when we encounter the last entry
//     (the last entry is recognized by de + de->rec_len > end 
//      or easier de->d_namlen == 0, see dir/codadir.c, DIR_Convert)  

NTSTATUS coda_readdir(IN FS_FILE_HANDLE DirectoryHandle,
                      IN OUT PULONG DirectoryOffset,
                      IN PFS_DIRECTORY_ENTRY OutputBuffer,
                      IN ULONG OutputBufferSize)
{
    coda_filehandle *dfh = (coda_filehandle *) DirectoryHandle;
    coda_inode *dir = dfh->fh_inode;

    NTSTATUS status;
    ULONG BytesRead;
    ULONG i;
    ULONG EntriesThatWillFit = 0;
    PFS_DIRECTORY_ENTRY Entry = OutputBuffer;
    ULONG Size;
    char *tmpbuf;
    PMDL pmdl;
    
    
    CDEBUG ((D_WRAPPER, "coda_readdir: Starting (fh %d: %d/%d)\n",
	     dfh->fh_unique, *DirectoryOffset, OutputBufferSize));

    LockInode(dir);
    
    try {
	
	//
	// First, do some sanity checking to be sure that the handle that
	// was passed in is really a valid directory handle.  For us, this means that
	// it is a pointer to a FILE and marked with TYPE directory.
	//
	if (!CODA_ISOURS(dir) ) {
	    CDEBUG((D_WRAPPER, "coda_readdir: invalid inode!\n"));
	    return(STATUS_INVALID_PARAMETER_1);
	}
	
	if (!CODA_ISDIR(dir) ) {
	    CDEBUG((D_WRAPPER, "coda_readdir: not a directory!\n"));
	    return(STATUS_INVALID_PARAMETER_1);
	}
	
	__try {
	    RtlZeroMemory(OutputBuffer, OutputBufferSize);
	}
	
	except(EXCEPTION_EXECUTE_HANDLER) {
	    
	    CDEBUG((D_WRAPPER, "coda_readdir: Coda readdir can't "
		   "clean buffer!\n"));
	    
	    return(STATUS_INVALID_USER_BUFFER);
	}
	
	
	//
	// Make a worst case determination of the number of entries 
	// that we can read at one time.
	//
	EntriesThatWillFit = OutputBufferSize
	    / (sizeof(FS_DIRECTORY_ENTRY)
	       + CODA_MAXNAMLEN * sizeof(WCHAR));
	
	CDEBUG((D_WRAPPER, "coda_readdir: %d Entries will fit.\n",
	       EntriesThatWillFit));
	
	tmpbuf = coda_malloc(OutputBufferSize);
	if ( !tmpbuf ) {
	    CDEBUG((D_WRAPPER,
		   "coda_readdir: Cannot allocate tmpbuf!\n"));
	    return STATUS_INSUFFICIENT_RESOURCES;
	}
	
	pmdl = MmCreateMdl(NULL, tmpbuf, OutputBufferSize);
	if ( !pmdl ) {
	    CDEBUG((D_WRAPPER,
		   "coda_readdir: Cannot allocate MDL\n"));
	    return STATUS_INSUFFICIENT_RESOURCES;
	}
	MmBuildMdlForNonPagedPool(pmdl);
	
	status = coda_read(dfh,
			   RtlConvertUlongToLargeInteger(*DirectoryOffset), 
			   pmdl, OutputBufferSize, &BytesRead);
	
	if(!NT_SUCCESS(status))	{
	    if (status == STATUS_END_OF_FILE) {
		CDEBUG((D_WRAPPER, "coda_readdir: end of directory\n"));
		return STATUS_NO_MORE_FILES;
	    }
	    CDEBUG((D_WRAPPER,
		    "coda_readdir: coda_read failed with %x\n", status));
	    return status;
	}
	
	CDEBUG((D_WRAPPER, "coda_readdir: Read %d bytes from container "
	       "file directory\n", BytesRead));

	status = coda_readdir_rewrite(dir, DirectoryOffset, tmpbuf,
				      BytesRead, OutputBuffer,
				      OutputBufferSize,
				      EntriesThatWillFit);
	
	//  if we didn't read any entries tell the wrapper
	if ( status == 0) {
	    CDEBUG((D_WRAPPER, "coda_readdir: end of directory\n"));
	    return STATUS_NO_MORE_FILES;
	}
	
	//  return bad errors
	if ( status < 0 ) {
	    CDEBUG((D_WRAPPER, "coda_readdir: returning error %d\n", -status));
	    return -status;
	}
	
	CDEBUG((D_WRAPPER, "coda_readdir: Returning Offset = %d, "
		"entries read = %d\n", *DirectoryOffset, status));
    } 
    
    finally {
	if (tmpbuf)
	    coda_free(tmpbuf);
	UnLockInode(dir);
    }
    
    EXIT(D_WRAPPER, "coda_readdir");
    return(STATUS_SUCCESS);
}


//
// Do Nothing - this is used to make sure coda_remshare gets called on close().
//

NTSTATUS coda_updshare (IN FS_FILE_HANDLE FileHandle, IN ULONG Access)
{
    coda_filehandle *fh = (coda_filehandle *) FileHandle;    
    CDEBUG ((D_WRAPPER, "coda_updshare: fh %d Access %x %s\n", fh->fh_unique,
	     Access, (fh->fh_isopen ? "is open" : "")));

    return  STATUS_SUCCESS;
}

// We are using this as a "Close" operation.  It appears to be
// called at every file close.

NTSTATUS coda_remshare (IN FS_FILE_HANDLE FileHandle, IN ULONG Access)
{
    coda_filehandle *fh = (coda_filehandle *) FileHandle;
    coda_inode *inode = fh->fh_inode;
    struct coda_ntvolume *ntvol;
    NTSTATUS rc;
    int flags = 0;

    CDEBUG ((D_WRAPPER, "coda_remshare: fh %d Access %x %s\n", fh->fh_unique,
	     Access, (fh->fh_isopen ? "is open" : "")));

    if ( !CODA_ISOURS(inode) ) {
	CDEBUG((D_WRAPPER, "coda_remshare: Bad file %s\n"));
	return STATUS_SUCCESS;
    }
    
    if ( !CODA_ISDIR(inode) && fh->fh_isopen
	 && (fh->fh_openflags & C_O_WRITE)) {
	// Forces writes to happen
	
	rc = OwFlushCache (FileHandle, 0);
	if (rc != STATUS_SUCCESS) {
	    CDEBUG((D_WRAPPER,
		    "coda_remshare: OwFlushCache returns error %d.\n",
		    rc));
	}
    }
    
    // Now "close" the file.
    
    if (fh->fh_isopen) {
	
	CDEBUG ((D_WRAPPER, "coda_remshare: Closing fh %d\n", fh->fh_unique));
	
	LockInode (inode);
	
	if (!(--(inode->c_open_cnt))) {
	    CDEBUG((D_FILE, "coda_remshare: Closing inode file object.\n")); 
	    // Close the object!
	    if (inode->c_object != NULL) {	
		inode_close_handle(inode);
	    } else {
		DbgPrint("coda_remshare: Closing NULL c_object\n");
	    }
	}
	
	rc = venus_close (&inode->c_fid, fh->fh_openflags, fh->fh_uid);
	
	// ignore error
	if ( rc ) {
	    DbgPrint("coda_remshare: venus_close error %d\n", rc);
	}
	
	// Update open flags
	fh->fh_openflags = 0;
	fh->fh_isopen = 0;
	UnLockInode(inode);
    }
    
    EXIT(D_WRAPPER, "coda_remshare");
    return STATUS_SUCCESS;
}


//
// coda_create
//
//   This routine is similar to lookup, except that the specified mode
//   for the file is to create it, rather than open it.
//
// Inputs:
//   DirectoryHandle - the directory where this file is being created.
//   FileName - the file to create
//   ShortFileName - unused (short name generated by FSDK.)
//   FileType - file/directory being created
//
// Outputs:
//   NewFileHandle - the newly created file handle.
//
// Returns:
//    SUCCESS - the operation was successful.
//    OBJECT_NAME_EXISTS - the specified name already exists.
//
// Notes:
//   None.
//


NTSTATUS coda_create (IN FS_FILE_HANDLE DirectoryHandle,
		      IN PWSTR FileName,
		      IN PWSTR ShortFileName,
		      IN ULONG FileType,
		      OUT PFS_FILE_HANDLE NewFileHandle)
     
{
    NTSTATUS status = STATUS_SUCCESS;
    UNICODE_STRING unicodeFileName;
    ANSI_STRING AnsiFileName;

    HANDLE fileHandle;
    PFILE_OBJECT fileObject;
    FS_FILE_ATTRIBUTES  tattr;  // Temp variable for getting attributes.
    
    CHAR NormalFileName[CODA_MAXNAMLEN];	
    coda_filehandle *dfh = (coda_filehandle *) DirectoryHandle;
    coda_filehandle *nfh ;
    coda_inode *dirinode = dfh->fh_inode;
    coda_inode *newinode;
    struct coda_vattr attrs;
    CodaFid newfid;
    int rc;	

    CDEBUG ((D_WRAPPER, "coda_create: dir fh %d\n", dfh->fh_unique));
    
    //
    // First, sanity checking
    //
    
    //
    // Sanity checking
    //
    if ( !CODA_ISOURS(dirinode) ) {
	DbgPrint("Bad file %s\n");
	return STATUS_INVALID_PARAMETER_1;
    }

    if (!CODA_ISDIR(dirinode)) {
	return(STATUS_NOT_A_DIRECTORY);
    }

    LockInode (dirinode);

    //
    // Now, build a unicode string for the file name.
    //

    unicodeFileName.Buffer = FileName;
    unicodeFileName.Length = wcslen(FileName) * sizeof(WCHAR);
    unicodeFileName.MaximumLength = unicodeFileName.Length + sizeof(WCHAR);
    RtlUnicodeStringToAnsiString(&AnsiFileName, &unicodeFileName, TRUE);
    if(AnsiFileName.Length > CODA_MAXNAMLEN) {
	UnLockInode (dirinode);
	return(STATUS_INVALID_PARAMETER_2);
    }
    memcpy(NormalFileName, AnsiFileName.Buffer, AnsiFileName.Length);
    NormalFileName[AnsiFileName.Length] = 0x0;
    RtlFreeAnsiString(&AnsiFileName);
    
    //
    // Tell venus to create the file.  Don't open the container until
    // needed.
    //
    
    if ( FileType != OW_HANDLE_TYPE_DIRECTORY ) {
	CDEBUG ((D_WRAPPER, "coda_create: create file %s\n", NormalFileName));

	rc = venus_create(&dirinode->c_fid, NormalFileName,
			  strlen(NormalFileName), 1, 0666, 0, 
			  &newfid, &attrs, dfh->fh_uid);
    } else {
	CDEBUG ((D_WRAPPER, "coda_create: create dir %s\n", NormalFileName));

	attrs.va_mode = 0777;
	rc = venus_mkdir(&dirinode->c_fid, NormalFileName,
			 strlen(NormalFileName), &newfid, &attrs, dfh->fh_uid);
    }
    

    if ( rc ) {
	UnLockInode (dirinode);
	return STATUS_ACCESS_DENIED;
    }
    
    // Now instantiate the new inode

    GrabLock();

    newinode =  inode_new(&newfid, dirinode, NormalFileName);
    if(!newinode) {
	CDEBUG((D_CNODE, "coda_create: inode_new failed!!!\n"));
	UnLockInode (dirinode);
	DropLock();
	return STATUS_INVALID_PARAMETER_1;
    }

    UnLockInode (dirinode);

    // Create the fh for the new file/directory

    nfh = get_cfh (newinode, dirinode);

    DropLock();

    if(!nfh) {
	CDEBUG((D_CNODE, "coda_create: get_cfh failed!!!\n"));
	return STATUS_INVALID_PARAMETER_1;
    }

    // Get the attributes
    rc = coda_getattr (nfh, &tattr);
    if (rc)
	DbgPrint ("coda_create: could not get attrs!\n");

    *NewFileHandle = (PFS_FILE_HANDLE)nfh;
  
    CDEBUG ((D_WRAPPER, "coda_create: returning fh %d\n", nfh->fh_unique));
    return STATUS_SUCCESS;
}



//
//
// coda_delete2 -- New version of delete
//

NTSTATUS coda_delete2(IN FS_FILE_HANDLE FileHandle,
		      IN FS_FILE_HANDLE ParentHandle,
		      IN PWSTR FileName)
{
    static delaydelete = 1;
    NTSTATUS Status;
    coda_filehandle *dfh = (coda_filehandle *) ParentHandle;
    coda_inode *dirinode = dfh->fh_inode;
    coda_filehandle *fh = (coda_filehandle *) FileHandle;
    coda_inode *finode = fh->fh_inode;

    int rc;
    
    CDEBUG ((D_WRAPPER, "coda_delete2: dir fh %d, fh %d\n",
	    dfh->fh_unique, fh->fh_unique));

    LockInode(dirinode);
    
    try {

	if (!CODA_ISOURS(finode)) {
	    return STATUS_INVALID_PARAMETER_1;
	}

	if (!CODA_ISOURS(dirinode)) {
	    return STATUS_INVALID_PARAMETER_2;
	}
	
	if (!CODA_ISDIR(dirinode)) {
	    return STATUS_NOT_A_DIRECTORY;
	}
	
	if (dirinode != finode->c_parent) {
	    return STATUS_CANNOT_DELETE;
	}

	// Inode contains name use it.  Wrapper can possibly
	// send NULL for FileName, so don't use it.
	
	LockInode(finode);
	
	if (!CODA_ISDIR(finode) && finode->c_fh_cnt > 1) {
	    // multiple file handles ... do an open to force
	    // the file to stay available .... 
	    // Just delete the directories!
	    char newname [CODA_MAXNAMLEN];
	    CDEBUG((D_WRAPPER, "coda_delete2: opening file...\n"));

	    rc = venus_open_by_path (&(finode->c_fid), C_O_WRITE, newname,
				     fh->fh_uid);

	    if (!rc) {
		inode_setflag(finode, I_DELETE);
		inode_cache_delete (finode);
	    } else {
		UnLockInode(finode);
		return STATUS_CANNOT_DELETE;
	    }
	}
	if (CODA_ISDIR(finode)) {
	    rc = venus_rmdir (&(dirinode->c_fid), finode->c_name,
			      strlen(finode->c_name), GetNtUID());
	} else {
	    rc = venus_remove (&(dirinode->c_fid), finode->c_name,
			       strlen(finode->c_name), GetNtUID());
	}

	if (rc)
	    Status = STATUS_CANNOT_DELETE;
	else
	    Status = STATUS_SUCCESS;

	UnLockInode(finode);

        if(NT_SUCCESS(Status) )
            {
		CDEBUG ((D_WRAPPER, "coda_delete2: calling coda_release\n"));
		inode_setflag(finode, I_PURGE);
		Status = coda_release(FileHandle);
            }
    } 

    finally {
        UnLockInode(dirinode);
    }
    
    EXIT (D_WRAPPER, "coda_delete2");
    return(Status);
}

//
// coda_rename2 ... rename a file
//

NTSTATUS coda_rename2(IN FS_FILE_HANDLE SourceDir,
		      IN FS_FILE_HANDLE TargetDir,
		      IN FS_FILE_HANDLE SourceHandle,
		      IN PWSTR SourceName,
		      IN PWSTR TargetName,
		      IN PWSTR AltTargetName,
		      OUT PFS_FILE_HANDLE TargetHandle)
{
    coda_filehandle *srcdfh = (coda_filehandle *) SourceDir;
    coda_inode *srcdir = srcdfh->fh_inode;
    coda_filehandle *trgdfh = (coda_filehandle *) TargetDir;
    coda_inode *trgdir = trgdfh->fh_inode;
    coda_filehandle *srcfh = (coda_filehandle *) SourceHandle;
    coda_inode *srcfile = srcfh->fh_inode;
    //    coda_filehandle *trgfh;
    //    coda_inode *trgfile;

    char srcName [CODA_MAXNAMLEN];
    char trgName [CODA_MAXNAMLEN];

    UNICODE_STRING unicodeFileName;
    ANSI_STRING AnsiFileName;

    int rc;

    CDEBUG ((D_WRAPPER, "coda_rename: src dfh %d, trg dfh %d, fh %d\n",
	     srcdfh->fh_unique, trgdfh->fh_unique, srcfh->fh_unique));
   
    if (!CODA_ISOURS(srcdir)) {
	return STATUS_INVALID_PARAMETER_1;
    }
    
    if (!CODA_ISOURS(trgdir)) {
	return STATUS_INVALID_PARAMETER_2;
    }

    if (!CODA_ISOURS(srcfile) || srcfile->c_parent != srcdir) {
	return STATUS_INVALID_PARAMETER_3;
    }

    // Lock Things!
    LockInode (srcdir);
    LockInode (trgdir);
    LockInode (srcfile);

    try {
	// Make the strings.....
	unicodeFileName.Buffer = SourceName;
	unicodeFileName.Length = wcslen(SourceName) * sizeof(WCHAR);
	unicodeFileName.MaximumLength = unicodeFileName.Length + sizeof(WCHAR);
	RtlUnicodeStringToAnsiString(&AnsiFileName, &unicodeFileName, TRUE);
	if(AnsiFileName.Length > CODA_MAXNAMLEN) {
	    return(STATUS_INVALID_PARAMETER_2);
	}
	memcpy(srcName, AnsiFileName.Buffer, AnsiFileName.Length);
	srcName[AnsiFileName.Length] = 0x0;
	RtlFreeAnsiString(&AnsiFileName);
	
	unicodeFileName.Buffer = TargetName;
	unicodeFileName.Length = wcslen(TargetName) * sizeof(WCHAR);
	unicodeFileName.MaximumLength = unicodeFileName.Length + sizeof(WCHAR);
	RtlUnicodeStringToAnsiString(&AnsiFileName, &unicodeFileName, TRUE);
	if(AnsiFileName.Length > CODA_MAXNAMLEN) {
	    return(STATUS_INVALID_PARAMETER_2);
	}
	memcpy(trgName, AnsiFileName.Buffer, AnsiFileName.Length);
	trgName[AnsiFileName.Length] = 0x0;
	RtlFreeAnsiString(&AnsiFileName);
	
	rc = venus_rename (&(srcdir->c_fid), srcName, strlen(srcName),
			   &(trgdir->c_fid), trgName, strlen(trgName),
			   srcdfh->fh_uid);
	
	if (rc) {
	    return STATUS_INVALID_HANDLE;
	}

	// inode manipulations ....
	inode_rename (srcfile, trgdir, trgName);
	
	*TargetHandle = (FS_FILE_HANDLE) srcfh;
    }

    finally {
	UnLockInode (srcdir);
	UnLockInode (trgdir);
	UnLockInode (srcfile);
    }

    EXIT (D_WRAPPER, "coda_rename");

    return STATUS_SUCCESS;
}


//
// coda_access -- this appears where the coda security must be handled!
//

NTSTATUS coda_access(IN FS_FILE_HANDLE FileHandle,
		     IN OUT PACCESS_STATE AccessState)
{
    coda_filehandle *fh = (coda_filehandle *) FileHandle;
    coda_inode *inode = fh->fh_inode;
    int flags;
    int i;
    ULONG mask;
    int rc;

    CDEBUG ((D_WRAPPER, "coda_access: name = '%s' %s fh %d, Prev = 0x%x, "
	     "Remain = 0x%x fh_granted = 0x%x\n", inode->c_name,
	     coda_f2s(&inode->c_fid), fh->fh_unique,
	     AccessState->PreviouslyGrantedAccess,
	     AccessState->RemainingDesiredAccess, fh->fh_granted));

    // Checking for now ... Should we do something?
    if (fh->fh_isopen) {
	CDEBUG ((D_WRAPPER, "coda_access: FH is open.  No action done.\n"));
    }

    // check for root directory and set no caching
    if (!root_init && fh == root_fh) {
	if (OwDisableFileCache(FileHandle,TRUE) != STATUS_SUCCESS)
	    CDEBUG((D_WRAPPER,
		    "coda_access: could not flush root dir cache\n"));
	root_init = TRUE;
    }

    // check access
    if ((fh->fh_granted & AccessState->RemainingDesiredAccess) ==
	AccessState->RemainingDesiredAccess) {

	CDEBUG ((D_WRAPPER, "coda_access: already has that access.\n"));

	AccessState->PreviouslyGrantedAccess |= 
	    AccessState->RemainingDesiredAccess;
	AccessState->RemainingDesiredAccess = 0;
	fh->fh_lastacc = AccessState->PreviouslyGrantedAccess;
        return STATUS_SUCCESS;	
    }

    LockInode (inode);

    if (AccessState->RemainingDesiredAccess) {
	
	// Create a coda flags with the proper requests.
	flags = 0;
	
	if (AccessState->RemainingDesiredAccess
	    & (FILE_READ_DATA | FILE_LIST_DIRECTORY | FILE_READ_ATTRIBUTES
	       | GENERIC_READ))
	    flags |= C_A_R_OK;
	if (AccessState->RemainingDesiredAccess
	    & (FILE_WRITE_DATA | FILE_APPEND_DATA | FILE_WRITE_ATTRIBUTES
	       | FILE_DELETE_CHILD | GENERIC_WRITE ))
	    flags |= C_A_W_OK;
	if (AccessState->RemainingDesiredAccess &
	    (FILE_EXECUTE | GENERIC_EXECUTE))
	    flags |= C_A_X_OK; 
	if (AccessState->RemainingDesiredAccess
	    & (FILE_ADD_FILE | FILE_ADD_SUBDIRECTORY))
	    flags |= C_A_C_OK;

	if (fh != coda_nt_volume.v_ctlfh) {

	    CDEBUG ((D_UPCALL, "venus_access: flags=%x\n", flags));

	    rc = venus_access (&(inode->c_fid), flags, fh->fh_uid);
	
	    if (rc) {
		CDEBUG ((D_WRAPPER, "coda_access: returning access denied\n"));
		UnLockInode (inode);
		return STATUS_ACCESS_DENIED;
	    }
	}
	    
	mask = AccessState->RemainingDesiredAccess 
	    & (FILE_READ_DATA | FILE_LIST_DIRECTORY | FILE_READ_ATTRIBUTES
	       | FILE_GENERIC_READ | FILE_WRITE_DATA | FILE_APPEND_DATA
	       | FILE_WRITE_ATTRIBUTES | FILE_DELETE_CHILD | FILE_GENERIC_WRITE
	       | FILE_EXECUTE | FILE_GENERIC_EXECUTE | FILE_ADD_FILE
	       | FILE_ADD_SUBDIRECTORY);
	
	AccessState->PreviouslyGrantedAccess |=
	    AccessState->RemainingDesiredAccess & mask;
	
	AccessState->RemainingDesiredAccess &= ~mask;
    }

    // Only add to the granted access rights.
    fh->fh_granted |= AccessState->PreviouslyGrantedAccess;
    fh->fh_lastacc = fh->fh_granted;

    UnLockInode (inode);

    CDEBUG ((D_WRAPPER, "coda_access: returning access granted 0x%x\n",
	     fh->fh_granted));

    return STATUS_SUCCESS;
}


// Normally called only on the CTL handle.

NTSTATUS coda_ioctl(IN FS_FILE_HANDLE FileHandle, IN ULONG ControlCode,
		    IN PVOID InputBuffer, IN ULONG InputBufferSize,
		    IN PVOID OutputBuffer, IN ULONG OutputBufferSize,
		    OUT PULONG BytesWritten, OUT PULONG BytesNeeded)
{

    NTSTATUS Status = STATUS_SUCCESS;

    CDEBUG((D_WRAPPER, "coda_ioctl: Starting, code: %x\n", ControlCode));
    
    switch (ControlCode) {
	
    case CODA_FSCTL_PIOCTL: 
	/// 
	/// Pioctl interface.  
	///
	{
	    ULONG cmd;
	    void *buf;
	    ULONG bufsize;
	    ULONG outsize = OutputBufferSize;

	    // Verify it is the correct fh ...
	    if (FileHandle != coda_nt_volume.v_ctlfh) {
		CDEBUG ((D_WRAPPER, "coda_ioctl: pioctl on wrong fh\n"));
		Status = STATUS_INVALID_DEVICE_REQUEST;
		break;
	    }
 
	    bufsize = MAX(InputBufferSize, OutputBufferSize);
	    
	    Status = STATUS_INVALID_DEVICE_REQUEST;
	    buf = coda_malloc(bufsize);
	    if  ( !buf  ) {
		DbgPrint("No memory for request data.\n");
		break;
	    }
	    
	    RtlCopyMemory(buf, InputBuffer, InputBufferSize);
	    
	    // Do the pioctl!
	    Status = coda_pioctl ((struct PioctlData*) buf, InputBufferSize,
				  OutputBuffer, &outsize);
	    
	    if ( Status ) {
		DbgPrint("coda_fsctrl: coda_pioctl returns error %d\n",
			 Status);
	    }
	    *BytesWritten = outsize;
	    coda_free (buf);
	}
	break;
	
    default:
	DbgPrint("coda_fsctrl: Invalid request %d\n", ControlCode);
	Status = STATUS_INVALID_DEVICE_REQUEST;
    }
    
    EXIT(D_WRAPPER, "coda_ioctl");
    return Status;
}


//
// Pseudo wrapper call call ... actually called from a CodaFsCtrl ...
//   but in UNIX this is a VFS call .... and thus like a wrapper call!
//

NTSTATUS coda_pioctl (IN struct PioctlData *buf, IN ULONG bufsize,
		      IN PVOID outbuf, IN OUT PULONG poutsize)
{
    char *FilePath = ((char *)(buf)) + sizeof (struct PioctlData);
    int   PathLen = strlen(FilePath);
    CodaFid *pfid;
    coda_inode *inode;
    int rc;

    CDEBUG( (D_PIOCTL, "coda_pioctl: Information!\n"));
    CDEBUG( (D_PIOCTL, "    FilePath = '%s', PathLen = %d\n", FilePath,
	    PathLen));
    CDEBUG( (D_PIOCTL, "    cmd = 0x%x\n", buf->cmd));
    CDEBUG( (D_PIOCTL, "    in_size = %d  out_size = %d \n", buf->vi.in_size,
	    buf->vi.out_size));

    // Fill in the buf->vi.in with the correct pointer!
    buf->vi.in = (caddr_t) (FilePath + PathLen + 1);

    // Find the file by name ....  Need to turn \ -> / ???  May not.
    if (PathLen) {

	// Look it up!  XXX May need more than this! 
	rc = coda_internal_path_lookup (coda_nt_volume.v_rootinode, 
					FilePath, 0, &inode);
	if (rc != STATUS_SUCCESS) {
	    CDEBUG( (D_PIOCTL,
		    "coda_pioctl: coda_internal_path_lookup error\n"));
	    *poutsize = 0;
	    return STATUS_INVALID_DEVICE_REQUEST;
	}
    } else {
	// Use the root fid!
	inode = coda_nt_volume.v_rootinode;
    }

    // Ask venus

    rc = venus_pioctl (&(inode->c_fid), buf, bufsize, outbuf, poutsize,
		       GetNtUID());

    CDEBUG( (D_PIOCTL, "coda_pioctl: returning, venus_pioctl rc = %d\n", rc));

    if (rc)
	return STATUS_INVALID_DEVICE_REQUEST;

    return STATUS_SUCCESS;
}

